@@ -1,3 +1,4 @@
+# frozen_string_literal: false
 require 'test/unit'
 require 'tempfile'
 
@@ -19,8 +20,8 @@ class TestException < Test::Unit::TestCase
       if bad
         bad = false
         retry
-        assert(false)
       end
+      assert(!bad)
     end
     assert(true)
   end
@@ -312,6 +313,10 @@ class TestException < Test::Unit::TestCase
     assert_raise_with_message(TypeError, /C\u{4032}/) do
       [*o]
     end
+    obj = eval("class C\u{1f5ff}; self; end").new
+    assert_raise_with_message(TypeError, /C\u{1f5ff}/) do
+      Class.new {include obj}
+    end
   end
 
   def test_errat
@@ -347,7 +352,7 @@ class TestException < Test::Unit::TestCase
   end
 
   def test_thread_signal_location
-    _, stderr, _ = EnvUtil.invoke_ruby("--disable-gems -d", <<-RUBY, false, true)
+    _, stderr, _ = EnvUtil.invoke_ruby(%w"--disable-gems -d", <<-RUBY, false, true)
 Thread.start do
   begin
     Process.kill(:INT, $$)
@@ -435,7 +440,7 @@ end.join
     bug3237 = '[ruby-core:29948]'
     str = "\u2600"
     id = :"\u2604"
-    msg = "undefined method `#{id}' for #{str.inspect}:String"
+    msg = "undefined method `#{id}' for \"#{str}\":String"
     assert_raise_with_message(NoMethodError, msg, bug3237) do
       str.__send__(id)
     end
@@ -629,6 +634,7 @@ end.join
   end
 
   def test_cause_raised_in_rescue
+    a = nil
     e = assert_raise_with_message(RuntimeError, 'b') {
       begin
         raise 'a'
@@ -636,6 +642,7 @@ end.join
         begin
           raise 'b'
         rescue => b
+          assert_same(a, b.cause)
           begin
             raise 'c'
           rescue
@@ -644,15 +651,17 @@ end.join
         end
       end
     }
-    assert_equal('a', e.cause.message, 'cause should not be overwritten by reraise')
+    assert_same(a, e.cause, 'cause should not be overwritten by reraise')
   end
 
   def test_cause_at_raised
+    a = nil
     e = assert_raise_with_message(RuntimeError, 'b') {
       begin
         raise 'a'
       rescue => a
         b = RuntimeError.new('b')
+        assert_nil(b.cause)
         begin
           raise 'c'
         rescue
@@ -661,6 +670,7 @@ end.join
       end
     }
     assert_equal('c', e.cause.message, 'cause should be the exception at raised')
+    assert_same(a, e.cause.cause)
   end
 
   def test_raise_with_cause
@@ -685,6 +695,7 @@ end.join
         begin
           raise 'b'
         rescue => b
+          assert_same(a, b.cause)
           begin
             raise 'c'
           rescue
@@ -694,6 +705,7 @@ end.join
       end
     }
     assert_equal('d', e.cause.message, 'cause option should be honored always')
+    assert_nil(e.cause.cause)
   end
 
   def test_cause_thread_no_cause
@@ -721,9 +733,9 @@ end.join
     bug12741 = '[ruby-core:77222] [Bug #12741]'
 
     x = Thread.current
-    a = false
+    q = Queue.new
     y = Thread.start do
-      Thread.pass until a
+      q.pop
       begin
         raise "caller's cause"
       rescue
@@ -735,9 +747,11 @@ end.join
       raise bug12741
     rescue
       e = assert_raise_with_message(RuntimeError, "stop") do
-        a = true
+        q.push(true)
         sleep 1
       end
+    ensure
+      y.join
     end
     assert_equal("caller's cause", e.cause.message)
   end
@@ -763,25 +777,121 @@ end.join
     assert_equal({}, e.arg, bug)
   end
 
+  def test_circular_cause
+    bug13043 = '[ruby-core:78688] [Bug #13043]'
+    begin
+      begin
+        raise "error 1"
+      ensure
+        orig_error = $!
+        begin
+          raise "error 2"
+        rescue => err
+          raise orig_error
+        end
+      end
+    rescue => x
+    end
+    assert_equal(orig_error, x)
+    assert_equal(orig_error, err.cause)
+    assert_nil(orig_error.cause, bug13043)
+  end
+
   def test_anonymous_message
     assert_in_out_err([], "raise Class.new(RuntimeError), 'foo'", [], /foo\n/)
   end
 
-  def test_name_error_info
-    obj = BasicObject.new
+  PrettyObject =
+    Class.new(BasicObject) do
+      alias object_id __id__
+      def pretty_inspect; "`obj'"; end
+      alias inspect pretty_inspect
+    end
+
+  def test_name_error_info_const
+    obj = PrettyObject.new
+
     e = assert_raise(NameError) {
       obj.instance_eval("Object")
     }
     assert_equal(:Object, e.name)
+
+    e = assert_raise(NameError) {
+      BasicObject::X
+    }
+    assert_same(BasicObject, e.receiver)
+    assert_equal(:X, e.name)
+  end
+
+  def test_name_error_info_method
+    obj = PrettyObject.new
+
     e = assert_raise(NameError) {
       obj.instance_eval {foo}
     }
     assert_equal(:foo, e.name)
+    assert_same(obj, e.receiver)
+
     e = assert_raise(NoMethodError) {
       obj.foo(1, 2)
     }
     assert_equal(:foo, e.name)
     assert_equal([1, 2], e.args)
+    assert_same(obj, e.receiver)
+    assert_not_predicate(e, :private_call?)
+
+    e = assert_raise(NoMethodError) {
+      obj.instance_eval {foo(1, 2)}
+    }
+    assert_equal(:foo, e.name)
+    assert_equal([1, 2], e.args)
+    assert_same(obj, e.receiver)
+    assert_predicate(e, :private_call?)
+  end
+
+  def test_name_error_info_local_variables
+    obj = PrettyObject.new
+    def obj.test(a, b=nil, *c, &d)
+      e = a
+      1.times {|f| g = foo; g}
+      e
+    end
+
+    e = assert_raise(NameError) {
+      obj.test(3)
+    }
+    assert_equal(:foo, e.name)
+    assert_same(obj, e.receiver)
+    assert_equal(%i[a b c d e f g], e.local_variables.sort)
+  end
+
+  def test_name_error_info_method_missing
+    obj = PrettyObject.new
+    def obj.method_missing(*)
+      super
+    end
+
+    e = assert_raise(NoMethodError) {
+      obj.foo(1, 2)
+    }
+    assert_equal(:foo, e.name)
+    assert_equal([1, 2], e.args)
+    assert_same(obj, e.receiver)
+    assert_not_predicate(e, :private_call?)
+
+    e = assert_raise(NoMethodError) {
+      obj.instance_eval {foo(1, 2)}
+    }
+    assert_equal(:foo, e.name)
+    assert_equal([1, 2], e.args)
+    assert_same(obj, e.receiver)
+    assert_predicate(e, :private_call?)
+  end
+
+  def test_name_error_info_parent_iseq_mark
+    assert_separately(['-', File.join(__dir__, 'bug-11928.rb')], <<-'end;')
+      -> {require ARGV[0]}.call
+    end;
   end
 
   def test_output_string_encoding
@@ -793,7 +903,7 @@ $stderr = $stdout; raise "\x82\xa0"') do |outs, errs, status|
       assert_equal 0, errs.size
       err = outs.first.force_encoding('utf-8')
       assert err.valid_encoding?, 'must be valid encoding'
-      assert_match /\u3042/, err
+      assert_match %r/\u3042/, err
     end
   end
 
@@ -816,6 +926,79 @@ $stderr = $stdout; raise "\x82\xa0"') do |outs, errs, status|
     assert_raise(NoMethodError, bug10969) {a.public_send("bar", true)}
   end
 
+  def test_message_of_name_error
+    assert_raise_with_message(NameError, /\Aundefined method `foo' for module `#<Module:.*>'$/) do
+      Module.new do
+        module_function :foo
+      end
+    end
+  end
+
+  def capture_warning_warn
+    verbose = $VERBOSE
+    warning = []
+
+    ::Warning.class_eval do
+      alias_method :warn2, :warn
+      remove_method :warn
+
+      define_method(:warn) do |str|
+        warning << str
+      end
+    end
+
+    $VERBOSE = true
+    yield
+
+    return warning
+  ensure
+    $VERBOSE = verbose
+
+    ::Warning.class_eval do
+      remove_method :warn
+      alias_method :warn, :warn2
+      remove_method :warn2
+    end
+  end
+
+  def test_warning_warn
+    warning = capture_warning_warn {@a}
+    assert_match(/instance variable @a not initialized/, warning[0])
+  end
+
+  def test_warning_warn_invalid_argument
+    assert_raise(TypeError) do
+      ::Warning.warn nil
+    end
+    assert_raise(TypeError) do
+      ::Warning.warn 1
+    end
+    assert_raise(Encoding::CompatibilityError) do
+      ::Warning.warn "\x00a\x00b\x00c".force_encoding("utf-16be")
+    end
+  end
+
+  def test_warning_warn_circular_require_backtrace
+    warning = nil
+    path = nil
+    Tempfile.create(%w[circular .rb]) do |t|
+      begin
+        path = File.realpath(t.path)
+        basename = File.basename(path)
+        t.puts "require '#{basename}'"
+        t.close
+        $LOAD_PATH.push(File.dirname(t))
+        warning = capture_warning_warn {require basename}
+      ensure
+        $LOAD_PATH.pop
+        $LOADED_FEATURES.delete(t)
+      end
+    end
+    assert_equal(1, warning.size)
+    assert_match(/circular require/, warning.first)
+    assert_match(/^\tfrom #{Regexp.escape(path)}:1:/, warning.first)
+  end
+
   def test_undefined_backtrace
     assert_separately([], "#{<<-"begin;"}\n#{<<-"end;"}")
     begin;
@@ -848,6 +1031,28 @@ $stderr = $stdout; raise "\x82\xa0"') do |outs, errs, status|
     end;
   end
 
+  def test_blocking_backtrace
+    assert_separately([], "#{<<~"begin;"}\n#{<<~'end;'}")
+    begin;
+      class Bug < RuntimeError
+        def backtrace
+          IO.readlines(IO::NULL)
+        end
+      end
+      bug = Bug.new '[ruby-core:85939] [Bug #14577]'
+      n = 10000
+      i = 0
+      n.times do
+        begin
+          raise bug
+        rescue Bug
+          i += 1
+        end
+      end
+      assert_equal(n, i)
+    end;
+  end
+
   def test_wrong_backtrace
     assert_separately([], "#{<<-"begin;"}\n#{<<-"end;"}")
     begin;
